#include "field.h"
#include <stdlib.h>
#include "math2d.h"
#include "debug.h"

FieldPoint * field;
Point fieldCenter;
int displayHeading;

// list of all masses
Mass * masses[MAX_MASSES];
// how many of them are in use?
int numMasses;

Point FieldToPoint(FieldPoint * p)
{
  Point res;
  res.y = (p - field) / FIELDSIZE;
  res.x = (p - field) % FIELDSIZE;
  return res;
}

void InitFieldMemory() 
{
  field = calloc(1, FIELDSIZE*FIELDSIZE*sizeof(FieldPoint));
  CleanField();
}

void FreeFieldMemory()
{
  free(field);
}

void CleanField()
{
  FieldPoint * p;
  for (p=field; p<field+FIELDSIZE*FIELDSIZE; p++) {
    p->owner = NULL;
    p->height = 0;		/* note: height doesn't matter */
  }
  /* set the new point of interest... in fact it shouldn't matter at
     all where it is once the field is clean. */
  fieldCenter.x = 0;
  fieldCenter.y = 0;
}

void DebugCheckField()
{
  int x, y;
  for (y=0; y<FIELDSIZE; y++) {
    for (x=0; x<FIELDSIZE; x++) {
      FieldPoint * p;
      p = FieldXY(x, y);
      if (p->owner) {
	assert(IsValidMass(p->owner));
	assert(p->height >= 0);
	assert(Wrap_IsPointInRect(XYToPoint(x, y), p->owner->maxRect));
      }
    }
  }
  DebugLevel(3, "DebugChecksField() passed without errors.\n");
}

Mass ** GetMassesInRect(Rect r)
{
  /* OPTIMIZE: using a clever algorithm here could allow a lot many
     more masses being around. Perhaps some kind of binary tree,
     parallel to masses[] or instead of them... */
  Mass ** res;
  int numFound = 0;
  int i;
  res = malloc(sizeof(Mass*));
  res[0] = NULL;
  for (i=0; i<numMasses; i++) {
    Rect r2;
    Mass * m;
    /* FIXME: check whether dead needed? */
    m = masses[i];
    r2 = Wrap_SmallestCommonRect(r, m->maxRect);
    if (r2.w && r2.h) {
      res[numFound] = m;
      numFound++;
      res = realloc(res, (numFound+1)*sizeof(Mass*));
      res[numFound] = NULL;
    }
  }
  return res;
}

void InitMassesMemory()
{
  numMasses = 0;
  /* no need to init anything else. */
}

void FreeMassesMemory()
{
  /* TODO: free all masses, should automatically free all interfaces.*/
}

Mass * CreateMass()
{
  Mass * m;
  static int createID = 0;
  if (numMasses >= MAX_MASSES) {
    DebugLevel(1, "Maximum number of masses reached (%d).\n", MAX_MASSES);
    return NULL;
  }
    
  m = masses[numMasses++] = calloc(1, sizeof(Mass));
  /* calloc filled the memory with zeros */
  /* assign default values for non-zero fields */
  m->maxRect.h = 0; m->maxRect.w = 0;
  m->speed = 256;
  m->player = -1;
  /* no movement, but flattening */
  m->oldFlattening = 1;
  m->weakSHR = FLATTENING_SHR + 1;
  m->normalSHR = FLATTENING_SHR;
  m->strongSHR = FLATTENING_SHR - 1;
  m->flatteningSHR[LEFT] = m->normalSHR;
  m->flatteningSHR[RIGHT] = m->normalSHR;
  m->flatteningSHR[UP] = m->normalSHR;
  m->flatteningSHR[DOWN] = m->normalSHR;

  m->minBorderHeight = MIN_BORDER_HEIGHT;
  m->maxBorderHeight = MAX_BORDER_HEIGHT;

  /* create a interface to the NULL mass */
  m->interfaces = calloc(1, sizeof(MassInterface));
  /* FIXME: a hack to not initialize the NULL interface... perhaps
     better F*** this NULL interface... */
  m->interfaces->age = 1;
  m->lookDistance = 10;

  m->heading = randarea(0, 256);

  /* creating a truely empty mass is mostly not wanted */
  m->bufferMass = 12*HEIGHT_NORM;

  RGBSET(m->debugColor, 127 + (createID % 2) * 128, 127 + (createID/2 % 2) * 128, 127 + (createID/4 % 2) * 128);
  createID++;

  DebugLevel(2, "Mass %X created.\n", (int)m);
  return m;
}

void DestroyMass(Mass * m)
{
  int nr;
  Point p;
  for (nr=0; nr<numMasses && masses[nr] != m; nr++) ;
  assert(masses[nr] == m);
  /* check wether the mass has still pixels on the screen */
  for (p.y=m->maxRect.y; p.y<m->maxRect.y+m->maxRect.h; p.y++) {
    for (p.x=m->maxRect.x; p.x<m->maxRect.x+m->maxRect.w; p.x++) {
      if (FieldPoint(p)->owner == m) FieldPoint(p)->owner = NULL;
    }
  }
#ifdef DEBUG_HEAVY
  for (p.x=0; p.x<FIELDSIZE; p.x++) for (p.y=0; p.y<FIELDSIZE; p.y++) {
    assert(FieldPoint(p)->owner != m);
  }
#endif
  free(m->baseLineGrowth);
  free(m->cell);
  DebugLevel(2, "Mass %X destroyed.\n", (int)m);
  free(m);
  masses[nr] = masses[--numMasses];
}

int IsValidMass(Mass * m)
{
  int nr;
  for (nr=0; nr<numMasses; nr++) if (masses[nr] == m) return 1;
  return 0;
}

Rect Wrap_SmallestCommonRect(Rect r1, Rect r2) {
  /* OPTIMIZE! */
  Rect s1, s2, r;
  assert(r1.w < FIELDSIZE && r1.h < FIELDSIZE);
  assert(r2.w < FIELDSIZE && r2.h < FIELDSIZE);
  assert(r1.h >= 0 && r1.w >= 0);
  assert(r2.h >= 0 && r2.w >= 0);
  r1.x &= FIELDWRAP;
  r1.y &= FIELDWRAP;
  r2.x &= FIELDWRAP;
  r2.y &= FIELDWRAP;
  s1 = SmallestCommonRect(r1, r2);
  //MIN(r1.x, r2.x) += FIELDSIZE;
  if (r1.x < r2.x) {
    r1.x += FIELDSIZE;
  } else {
    r2.x += FIELDSIZE;
  }
  //MIN(r1.y, r2.y) += FIELDSIZE;
  if (r1.y < r2.y) {
    r1.y += FIELDSIZE;
  } else {
    r2.y += FIELDSIZE;
  }
  s2 = SmallestCommonRect(r1, r2);
  /* now unite s1 and s2 (aka "biggestCommonRect") */
  /* This actually could happen, but I think it won't. It would make
     things somewhat more complicated. */
  assert(r1.w + r2.w < FIELDSIZE);
  assert(r1.w + r2.w < FIELDSIZE);
  if (s1.w > 0) {
    /* This could acutally happen, but I think it won't... */
    assert(s2.w == 0);
    r.x = s1.x;
    r.w = s1.w;
  } else {
    r.x = s2.x;
    r.w = s2.w;
  }
  if (s1.h > 0) {
    /* This could acutally happen, but I think it won't... */
    assert(s2.h == 0);
    r.y = s1.y;
    r.h = s1.h;
  } else {
    r.y = s2.y;
    r.h = s2.h;
  }
  return r;
}

void Wrap_ExpandRectToIncludePoint(Rect * r, int x, int y)
{
  /* OPTIMIZE? But before, check everywhere if the non-wrap version
     would be enough */
  /* FIXME: re-read for correctness. */
  Rect r1, r2;
  /*r->w/h == FIELDSIZE could actually happen, but I think it won't. */
  assert(r->w < FIELDSIZE && r->h < FIELDSIZE);
  assert(r->h >= 0 && r->w >= 0);
  x &= FIELDWRAP;
  y &= FIELDWRAP;
  r->x &= FIELDWRAP;
  r->y &= FIELDWRAP;
  r1 = *r;
  r2 = *r;
  r2.x -= FIELDSIZE;
  r2.y -= FIELDSIZE;
  /* do it twice... */
  ExpandRectToIncludePoint(&r1, x, y);
  ExpandRectToIncludePoint(&r2, x, y);
  /* ...and pick the smaller one. */
  if (r1.w < r2.w) {
    r->w = r1.w;
    r->x = r1.x;
  } else {
    r->w = r2.w;
    r->x = r2.x;
  }
  if (r1.h < r2.h) {
    r->h = r1.h;
    r->y = r1.y;
  } else {
    r->h = r2.h;
    r->y = r2.y;
  }
}

int Wrap_GetCompareDistance(Point p1, Point p2)
{
  int dx, dy;
  p1.x &= FIELDWRAP;
  p1.y &= FIELDWRAP;
  p2.x &= FIELDWRAP;
  p2.y &= FIELDWRAP;

  dx = abs(p1.x-p2.x);
  dy = abs(p1.y-p2.y);
  /* there are no distances longer than FIELDSIZE/2 (on one axis) */
  if (dx > FIELDSIZE/2) dx = (FIELDSIZE - dx);
  if (dy > FIELDSIZE/2) dy = (FIELDSIZE - dy);
  return sqr(dx) + sqr(dy);
}

double Wrap_GetRealDistance(Point p1, Point p2)
{
  return sqrt((double)(Wrap_GetCompareDistance(p1, p2)));
}

void Wrap_FitPointIntoRect(Point * p, Rect r)
{
  int dleft, dright;
  assert(r.w > 0 && r.h > 0);
  assert(r.w < FIELDSIZE);
  assert(r.h < FIELDSIZE);

  p->x &= FIELDWRAP;
  r.x &= FIELDWRAP;
  if (p->x < r.x) r.x -= FIELDSIZE;
  assert(p->x >= r.x);
  if (p->x >= r.x + r.w) {
    dright = p->x - (r.x + r.w - 1);
    dleft  = r.x - (p->x - FIELDSIZE);
    assert(dright > 0 && dleft > 0);
    if (dleft < dright) {
      p->x = r.x;
    } else {
      p->x = r.x + r.w - 1;
    }
  } 

  p->y &= FIELDWRAP;
  r.y &= FIELDWRAP;
  if (p->y < r.y) r.y -= FIELDSIZE;
  assert(p->y >= r.y);
  if (p->y >= r.y + r.h) {
    dright = p->y - (r.y + r.h - 1);
    dleft  = r.y - (p->y - FIELDSIZE);
    assert(dright > 0 && dleft > 0);
    if (dleft < dright) {
      p->y = r.y;
    } else {
      p->y = r.y + r.h - 1;
    }
  } 
}

int Wrap_IsPointInRect(Point p, Rect r)
{
  Point p2;
  p2 = p;
  Wrap_FitPointIntoRect(&p2, r);
  p.x &= FIELDWRAP;
  p.y &= FIELDWRAP;
  p2.x &= FIELDWRAP;
  p2.y &= FIELDWRAP;
  return p.x == p2.x && p.y == p2.y;
}

Rect GetThawnRect()
{
  Rect r;
  r.x = displayCenter.x - FIELDFREEZE/2;
  r.y = displayCenter.y - FIELDFREEZE/2;
  r.w = FIELDFREEZE;
  r.h = FIELDFREEZE;
  return r;
}

int randarea(int smallest, int size) {
  /* TODO: make sure it's a function that returns the same results on
     every system (in case the same start value was used).
     Will be needed for networking. */
  return rand() % size + smallest;
}

Point Wrap_VectorFromTo(Point from, Point to)
{
  /* TODO!!! */
  return VectorFromTo(from, to);
}
