/* 
   mass - a game without bitmaps

   Copyright (C) 2001 Martin Renold

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include "video2d.h"
#include "video3d.h"
#include "mass.h"
#include "field.h"
#include "game.h"
#include "debug.h"
#include "networkplayer.h"
#include "input.h"

/* for the timer */
#include "SDL.h"

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>

typedef enum { F_NONE = 0, F_2D, F_3D, F_WRAP } Frontend;

/* tries to start newFrontend, and returns the one actually started */
int StartFrontend(Frontend oldFrontend, Frontend newFrontend, int argc, char ** argv) {
  int i;
  static int inputStarted = 0;
  if (newFrontend == F_WRAP) newFrontend = 1;
  if (oldFrontend == newFrontend) return oldFrontend;
  if (oldFrontend == F_2D) DoneVideo2d();
  if (oldFrontend == F_3D) DoneVideo3d();
  if (newFrontend == F_NONE) return F_NONE;

  for (i=0; i<F_WRAP-1; i++) {
    if (i) DebugLevel(1, "Trying next frontend...\n");
    if (newFrontend == F_2D && !InitVideo2d(argc, argv)) {
      /* SDL needs video initialized before the keymap to read the
	 config file ist available :-( */
      if (!inputStarted) {
	inputStarted = 1;
	InitInput();
      }
      return newFrontend;
    }
    if (newFrontend == F_3D && !InitVideo3d(argc, argv)) {
      /* SDL needs video initialized before the keymap to read the
	 config file ist available :-( */
      if (!inputStarted) {
	inputStarted = 1;
	InitInput();
      }
      return newFrontend;
    }
    newFrontend++;
    if (newFrontend == F_WRAP) newFrontend = 1;
  }
  printf("Did not find any way to show the game, sorry :-(\n");
  return F_NONE;
}

void SetPlayerActionsFromKeyboard()
{
  localPlayer[0].movex = 0;
  localPlayer[0].movey = 0;
  if (key.player1left.pressed)  localPlayer[0].movex--;
  if (key.player1up.pressed)    localPlayer[0].movey--;
  if (key.player1right.pressed) localPlayer[0].movex++;
  if (key.player1down.pressed)  localPlayer[0].movey++;
  localPlayer[0].button1 = key.player1button1.pressed;
  localPlayer[0].button2 = key.player1button2.pressed;

  localPlayer[1].movex = 0;
  localPlayer[1].movey = 0;
  if (key.player2left.pressed)  localPlayer[1].movex--;
  if (key.player2up.pressed)    localPlayer[1].movey--;
  if (key.player2right.pressed) localPlayer[1].movex++;
  if (key.player2down.pressed)  localPlayer[1].movey++;
  localPlayer[1].button1 = key.player2button1.pressed;
  localPlayer[1].button2 = key.player2button2.pressed;
}

int main (int argc, char **argv)
{
  int i;
#define TIME_PRECISION 16	/* beware the overflow */
  Uint32 minFrameDuration = 0;
  Uint32 lastFrameTime; 	/* in miliseconds*TIME_PRECISION since SDL_Init */
  Uint32 lastShownFrameTime;

  Frontend frontend = F_2D;

  printf("mass version %s\n", VERSION);

  for (i=0; i<argc; i++) {
    if (!strcmp(argv[i], "--help") || !strcmp(argv[i], "-h")) {
      printf("Command line switches:

--game list        list possible game types
--grid             [OpenGL] draw grids only
--fps number       try to hold that much game frames per second (0 unlimited)
");
      exit(0);
    }
    if (!strcmp(argv[i], "--fps")) {
      int fps;
      if (i == argc) {
	printf("Usage: %s --fps VALUE\n", argv[0]);
	exit(1);
      }
      i++;
      fps = atoi(argv[i]);
      if (fps <= 0) {
	minFrameDuration = 0;
	DebugLevel(1, "No fixed FPS, always running at maximum CPU speed.\n");
      } else {
	minFrameDuration = (TIME_PRECISION*1000) / fps;
	DebugLevel(1, "Max FPS set to %d (which means min frame duration %d/%dms)\n", fps, minFrameDuration,TIME_PRECISION);
      }
    }
  }

#ifdef DEBUG
  SDL_Init(SDL_INIT_NOPARACHUTE);
#else
  SDL_Init(0);
#endif

  /* TODO: network random code --> engine. */
  srand(time(NULL));
  Assertions();

  InitFieldMemory();
  InitMassesMemory();

  InitNetworkPlayer();

  /* prepare the game field */
  StartGame(argc, argv);

  /* tick, tack. */
  lastFrameTime = lastShownFrameTime = TIME_PRECISION * SDL_GetTicks();

  /* pick and start a frontend to display the game */
  frontend = StartFrontend(F_NONE, frontend, argc, argv);

  while (frontend) {
    int framedrop = 0;
    /* get new keyboard and mouse state */
    if (key.pause.toggled) SDL_WaitEvent(NULL);
    UpdateInput();

    /* feed network and/or local loopback with keyboard data */
    /* TODO: allow input from logfiles instead of keyboard */
    SetPlayerActionsFromKeyboard();

    if (!key.pause.toggled || key.nextFrame.hit) {
      static int frameno=0;
      /* networkplayer.c; FIXME: move the frame counter away!! */
      UpdatePlayerInput(frameno++);

      /* delay the completion, or set framedrop */
      if (minFrameDuration) {
	Uint32 time;
	time = TIME_PRECISION * SDL_GetTicks();
	/* sleep if it is running too fast */
	while (minFrameDuration > time - lastFrameTime) {
	  SDL_Delay(minFrameDuration - (time - lastFrameTime));
	  time = TIME_PRECISION * SDL_GetTicks();
	}
	assert(time >= lastFrameTime + minFrameDuration);
	lastFrameTime += minFrameDuration;
	/* resync if it runs *way* too slow: */
	/* if it the engine lags behind > 200ms, give up to keep the
	   desired framerate. It might become better... */
	/* this also ensures that the screen gets refreshed even if
	   the calculations are lagging behind. TODO: This could be a
	   kind of "min-display-FPS" parameter */
	/* FIXME: this is nonsense for <= 2 FPS */
	if (lastFrameTime - lastShownFrameTime > TIME_PRECISION*200) {
	  static int timeDropped = 0;
	  timeDropped++;
	  if (timeDropped == 2) {
	    /* mplayer-style complaint :-) */
	    printf("
         ************************************************
         **** Your system is too SLOW to play this!  ****
         ************************************************
!!! Possible reasons, problems, workarounds: 
- The game has currently to calculate more masses than usual. You
  could simply wait until some of them disappear, or move away.
- Try to give a more realistic value to --fps, or use no FPS control
  at all (--fps 0).
- Think about it, review the code in main.c and fix all the FIXMEs.\n\n");
	    /* TODO: propose a FPS (--fps 22), or use a --framedrop
	       parameter to control maximum frame dropping... */
	    /* FIXME: this is not Frames Per Second, but Rounds Per
	       Second (RPS?). The FPS value is independant. */
	  }
	  lastFrameTime = time;
	}
	/* if the engine lags behind > 100ms, start dropping frames */
	if (time - lastFrameTime > TIME_PRECISION*100) {
	  static int framesDropped = 0;
	  framesDropped++;
	  if (framesDropped == 21) {
	    DebugLevel(1, "INFO: Displaying of more than 20 Frames dropped to hold the desired FPS.\n");
	  }
	  framedrop = 1;
	}
	if (!framedrop) lastShownFrameTime = lastFrameTime;
      }

      /* let the engine work */
      if (NextGameFrame()) {
	/* game over */
	frontend = StartFrontend(frontend, F_NONE, argc, argv);
      }
    }

    if (!framedrop) {
      if (frontend == F_2D) DrawFrame2d();
      if (frontend == F_3D) DrawFrame3d();
    }

    if (key.switchInterface.hit) {
      frontend = StartFrontend(frontend, frontend+1, argc, argv);
    }
    if (key.quit.hit) {
      frontend = StartFrontend(frontend, F_NONE, argc, argv);
    }
  }

  /* the frontend has already terminated at this point */

  FreeNetworkPlayer();

  FreeMassesMemory();
  FreeFieldMemory();

  SDL_Quit();
  exit(0);
}
