#ifndef __MASS__
#define __MASS__

#include <limits.h>
#include "math2d.h"

#define MAX_MASSES 1000

/* We should be able to sum up the height values of all pixels of a mass 
 * using an int (without overflow).  HEIGHT_MAX is not a "hard" limit, but
 * the average height  should be significant lower. 
 * (TODO: make it machine-independant so network play is synchronized) */
#define HEIGHT_MAX (INT_MAX / 256)
/* HEIGHT_NORM ist the height difference that will be shadowed as a
 * 45 degree slope (this means it is equal to the distance between two
 * neighbour pixels). */
#define HEIGHT_NORM (HEIGHT_MAX / 64)

/* Flattening:
 *
 * The flattening algo takes from the height of each pixel
 * 4*(height>>FLATTENING_SHR) away to distribute it amog the four
 * neighbour pixels. */
#define FLATTENING_SHR 3
/* In case the height of a border pixel is less than MIN_HEIGHT, 
 * the algo gives this pixel up. */
#define MIN_BORDER_HEIGHT (HEIGHT_NORM)
/* In case the height of a border pixel is more than MAX_HEIGHT,
 * the algo tries to create a new pixel next to it. */
#define MAX_BORDER_HEIGHT (HEIGHT_NORM*5)
/* see struct _Mass_ */
#define MAX_SHAPEPOINTS 16

typedef struct _Mass_ Mass;
typedef struct _MassInterface_ MassInterface;

/* Used in the mass below. See PreFlattening_ShapePoints() in mass.c
   for details */
typedef struct {
  Point pos;
  int radius;   // 0 = not used, 5 = 5 pixels radius
  int strength; // 0 = not used, <0 = distract, >0 = attract the mass
} ShapePoint;

struct _MassInterface_ {
  MassInterface * next;
  /* the target mass, or NULL for the interface to all empty points */
  Mass * m;
  /* Rounds alive (relative to the parent masses's age) */
  int age;
  /* number of pixels currently touching each other */
  int contact, oldContact;
  /* 0: stay; 128 half-join; 256 join equal; -256 shrink equal */
  int borderReaction;
  unsigned eatme : 1;
};

typedef struct {
  /* The BaseLineGrowth algo draws a single line through the mass.
     The line goes through the mass center. Start and end point are
     threated specially. */
  /* How much height to add to each pixel on the line */
  int height;
  /* Extra-height to add to start / endpoint. */
  int start, end;
  /* distance between the pixel of the line (256 = 1 pixel) */
  int step;
  /* start/end of the mass; calculated by the algo */
  Point head, tail;
  /* if set, the first step is not always in the mass center, but
     still on the line. */
  unsigned randomStepStart : 1;
  /* if set, the height is not added but re-set (if > 0). Note that
     the varibles start / end are used for the endpoints. */
  unsigned setHeight : 1;
} BaseLineGrowth;

typedef struct {
  /* FIXME: much too much states. */
  /* a trial to implement game of life with masses ;-) */
  /* game of life state - 0 or 1 */
  int golState;
  /* direction where this being will be going, perhaps. */
  int wishDir;
  /* kind of activation */
  int pulse;
} CellInfo;

/* This represents one "physical" mass on the screen. */
struct _Mass_ {
  /* the smallest rectangle that encloses the mass on the screen */
  Rect maxRect; 
  /* the height to be added to the mass as a whole */
  int bufferMass;
  /* how many pixels? should always remain accurate  */
  int numPixels;
  /* sum of all height values; not accurate, auto-updated */
  int volume;
  /* for move backward, no movement, move forward */
  int weakSHR, normalSHR, strongSHR;
  /* controls the strength of the flattening algo [LEFT][UP]... */
  int flatteningSHR[4];
  /* belongs to the flattening algo too */
  int minBorderHeight, maxBorderHeight;
  /* how often the mass is processed per frame */
  int speed;
  int pos;
  /* how the mass interacts with the masses nearby (never NULL) */
  MassInterface * interfaces;
  /* maximum distance for interfaces masses from the maxRect; needs to
     be greater than the maximum pixel movement per round of this mass */
  int lookDistance;
  /* basic color */
  Color baseColor;
  /* for debugging only */
  Color debugColor;
  /* for the frontend that displays the mass */
  enum {
    DT_NORMAL = 0, DT_FLAT
  } displayType;
  /* these points form the shape of the mass */
  ShapePoint shapePoints[MAX_SHAPEPOINTS];
  /* generally the center of the rectangle (FIXME: not yet used.) */
  /*Point centerpoint;*/
  /* (temporary) how many rounds the mass will grow */
  int growRounds;
  /* the direction, if any. 0..255 */
  int heading;
  /* m->speed is how often the mass is calculated, this is movement */
  int velocity;
  /* player id controling this mass, -1 = if it isn't a player mass */
  int player;
  /* automatically increases, but might be reset */
  int age;
  /* general energy counter */
  int energy;
  /* how vivid the mass is; causes random movements in it */
  int vivid;
  /* same with avoid points (experiment) */
  int vivid2;
  /* create a new fruit when 256 */
  int plantGrowFruit;
  /* this algo grows mass around a line */
  BaseLineGrowth * baseLineGrowth;
  /* this algo behaves like a cell, getting input from the neighbours */
  CellInfo * cell;
  /* how near to the nexus is the mass? */
  int nexusLevel;
  /* was the mass changed? */
  unsigned mustRedraw : 1;
  /* move according to m->heading and m->velocity, or just flatten? */
  unsigned flatteningMovement : 1;
  /* do the flattening with the convolution matrix algo */
  unsigned convolveFlattening : 1;
  /* do the flattening with the old (fast but weak) algo */
  unsigned oldFlattening : 1;
  /* can a player simply eat this mass? */
  unsigned playerFood : 1;
  unsigned isPlant : 1;
  unsigned isPlant2 : 1;
  unsigned isBomb : 1;
  unsigned isWall : 1;
  unsigned isProjectil : 1;
  unsigned isMeltingMass : 1;
  unsigned isMazeWall : 1;
  unsigned isCoinMass : 1;
  /* grow much much faster (not linear) */
  unsigned superGrow : 1;
};

/*
 * for main()
 */
extern void ContinueMasses();

/* create a new mass touching the field */
extern Mass * CreateMass();
/* create a mass and take a free starting pixel near (x, y). May
   return NULL. */
extern Mass * CreateMassNearby(int x, int y, int radius);

/* Destroy a mass. It will be cleanly removed immediately. */
extern void DestroyMass(Mass * m);

/* possible reactions to a border point conflict with another mass */
typedef enum { 
  BP_ERROR = -1, 
  BP_SHRINK = -2, 
  BP_EXPAND = -3, 
  BP_STAY = 0,
  BP_HALFJOIN = 128, 		/* any value between is OK */
  BP_JOIN = 256			/* biger values are OK */
} BP_Reaction;

/* get interface between 2 masses, create one if none */
extern MassInterface * CreateInterface(Mass * m1, Mass * m2);
/* get interface between 2 masses, NULL if none */
extern MassInterface * GetInterface(Mass * m1, Mass * m2);
/* delete it */
extern void DeleteInterface(MassInterface * mi);

/* looks for a (p->owner == NULL) pixel at the border of the mass and adds it */
extern Point AddBorderPixel(Mass * m);

extern void DebugCheckMass(Mass * m);
extern void DebugPrintMass(Mass * m);

#endif

