/* There is no network support implemented yet. I plan using SDL_Net.
 * 
 * * client/server vs peer to peer
 *
 * P2p is best for only a few players. It minimizes the lag there, but
 * makes it hard to drop a player in case his connection fails,
 * because no client knows whether the other peers got his data.
 *
 * Client/server minimizes the traffic for the clients, but adds
 * somewhat lag overhead. The plus is that the server may simply
 * decide not to update a laggy player in order to keep the game for
 * the others running.
 *
 * * tpc vs udp
 * 
 * Udp seems the better choice, as it will not block in case a packet
 * got lost. The trick is to always send redundant information, so the
 * clients will hopefully get all data even if a packet was lost. In
 * the worst case the a client is missing some player's data and has
 * to ask for it.
 *
 * Tcp has the advantages that there are no problems with NAT
 * firewalls, and packet loss is handled automagically (but
 * slowly). Perhaps it should be supported as fallback.
 *
 * For more detailed discussion, look at the FreeCraft network docs:
 * http://freecraft.sourceforge.net/doxygen/NetworkModule.html#NetworkModule 
 *
 * * How to join a network game?
 *
 * Most complicated but coolest would be, if the player starts a local
 * game, and once he starts moving out of the screen, a network
 * connection is attempted to any other player (currently in the
 * middle of gaming, too). Both game states are merged, and the
 * network join would go without any player noticing - until they
 * meet.
 *
 * Less complicated: when you start a game, you may specify how many
 * players may join it. To join a game, you have to download the
 * gamestate from the player and throw the own away.
 *
 * Simple version: Games have to be started simultanously, and
 * there's no need to send the state of a whole running game at all,
 * only game type random number generator have to be synced.
 */
#include "networkplayer.h"
#include "input.h"
#include "stdlib.h"
#include "debug.h"

#define PACKETQUEUE_MAX (MAX_PLAYERS * (MAX_LAG+1))

int numPlayers;
PlayerInput player[MAX_PLAYERS];
PlayerInput localPlayer[MAX_PLAYERS];
int lag = 0; 
int GetNetworkLag() { return lag; }

struct {
  int frame;
  int request;
  void * data;
} packetQueue[PACKETQUEUE_MAX];

typedef struct {
  int player; /* id is unique on all computers playing the same game */
  int movex; /* -1, 0, +1 */
  int movey;
  int button1; /* 0 or 1 */
  int button2;  
} REQ_PlayerInput;

enum { REQ_EMPTY, REQ_PLAYERINPUT };

int numPacketsQueued;

void InitNetworkPlayer()
{
  int i;
  numPlayers = 0;
  numPacketsQueued = 0;
  for (i=0; i<MAX_PLAYERS; i++) {
    player[i].exists = 0;
    localPlayer[i].exists = 0;
  }
}

void CleanupNetworkPlayer()
{
  int i;
  for (i=0; i<numPacketsQueued; i++) {
    free(packetQueue[i].data);
  }
  numPacketsQueued = 0;
  numPlayers = 0;
  /* FIXME: remove local and network players? */
}

void FreeNetworkPlayer()
{
  CleanupNetworkPlayer();
}

void AddLocalPlayer(int id)
{
  assert(id >= 0 && id < MAX_PLAYERS);
  assert(localPlayer[id].exists == 0);
  memset(&localPlayer[id], 0, sizeof(PlayerInput));
  localPlayer[id].exists = 1;
  localPlayer[id].id = numPlayers;
  /* FIXME: when two players connect within the same lag spawn or
     even at once, they might receive the same ID because numPlayers
     is not yet incrementet (and can't be on every host). For the
     moment, do not join too fast :-) */
  assert(numPlayers+1 < MAX_PLAYERS);
  assert(player[numPlayers].exists == 0);
}

void RemoveFromPacketQueue(int nr)
{
  free(packetQueue[nr].data);
  packetQueue[nr].data = NULL;
  packetQueue[nr].request = REQ_EMPTY;
}

void CleanPacketQueue()
{
  int writePos;
  int readPos;
  for (readPos=0,writePos=0; readPos < numPacketsQueued; readPos++,writePos++) {
    packetQueue[writePos] = packetQueue[readPos];
    if (packetQueue[readPos].request == REQ_EMPTY) {
      assert(!packetQueue[readPos].data);
      writePos--;
    }
  }
  numPacketsQueued = writePos;
}

int UpdatePlayerInput(int currentFrame)
{
  int nr, i;
  int data_from_how_many_players;
  REQ_PlayerInput * pin;
  /* 1. send keyboard data for all local players (for frame
     (currentFrame+lag)) */
  for (nr = 0; nr < MAX_PLAYERS; nr++) {
    if (localPlayer[nr].exists) {
      /* TODO: for each network connection, send data to host. 
         - do send all information in one single packet, to make sure
         either everything is lost or nothing, so it is not possible
         to miss a newly added player. */
      /* and last but not least: add it to the local queue */
      i = numPacketsQueued++;
      assert(numPacketsQueued != PACKETQUEUE_MAX);
      packetQueue[i].frame = currentFrame + lag;
      packetQueue[i].request = REQ_PLAYERINPUT;
      pin = malloc(sizeof(REQ_PlayerInput));
      pin->player = localPlayer[nr].id;	
      pin->movex = localPlayer[nr].movex;
      pin->movey = localPlayer[nr].movey;
      pin->button1 = localPlayer[nr].button1;
      pin->button2 = localPlayer[nr].button2;
      packetQueue[i].data = pin;
    }
  }

  /* 2. check the queue for the current frame and read missing data
     from the network */

  do { 
    data_from_how_many_players = 0;
    for (i=0; i<numPacketsQueued; i++) {
      if (packetQueue[i].frame == currentFrame && packetQueue[i].request == REQ_PLAYERINPUT) {
	data_from_how_many_players++;
      }
    }
    if (data_from_how_many_players < numPlayers) {
      /* TODO: wait for (more) network packets */
      /* hint: seperate function */
      assert(0);
    }
  } while (data_from_how_many_players < numPlayers);
  /* data from every known host is in the queue, now read it. */
  for (i=0; i<numPacketsQueued; i++) {
    if (packetQueue[i].frame == currentFrame && packetQueue[i].request == REQ_PLAYERINPUT) {
      pin = packetQueue[i].data;
      if (pin->player >= numPlayers) {
	/* a new player! */
	assert(pin->player == numPlayers);
	assert(pin->player >= 0 && pin->player < MAX_PLAYERS);
	assert(player[pin->player].exists == 0);
	memset(&player[pin->player], 0, sizeof(PlayerInput));
	localPlayer[pin->player].exists = 1;
      }
      player[pin->player].movex = pin->movex;
      player[pin->player].movey = pin->movey;
      player[pin->player].button1 = pin->button1;
      player[pin->player].button2 = pin->button2;
      RemoveFromPacketQueue(i);
    }
  }
  CleanPacketQueue();
  return 0;
}
