#include "video2d.h"
#include "field.h"
#include "debug.h"
#include "networkplayer.h"
#include "input.h"
#include "math2d.h"
#include "game.h"
#include "SDL.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/* globals */
SDL_Surface * screen;

#define DISPLAY_SCREENX 256
#define DISPLAY_SCREENY 256

/* the displayed screen rectangle, in mass coordinates */
Rect screenRect = {0,0,DISPLAY_SCREENX,DISPLAY_SCREENY};

/* clear the frame, set the background to black */
extern void ClearFrame();

/* r, g, b are unsigned chars, 255/255/255 is white */
#define SetPixel(x, y, rgb) (*((Uint16*)((char*)screen->pixels + (y)*screen->pitch) + (x)) = (((rgb[0])>>3) << 10) + (((rgb[1])>>3) << 5) + (((rgb[2])>>3) << 0) )
/* #define SetPixel(x, y, c) SetPixel(x, y, c.r, c.g, c.b) */
#define GetPixel(x, y, rgb) { \
  Uint16 GetPixel_pixel = *((Uint16*)((char*)screen->pixels + (y)*screen->pitch) + (x)); \
  rgb[0] = ( (GetPixel_pixel >> 10) & 31 ) << 3; \
  rgb[1] = ( (GetPixel_pixel >>  5) & 31 ) << 3; \
  rgb[2] = ( (GetPixel_pixel >>  0) & 31 ) << 3; \
}
void InvertPixel(int x, int y) {
  int j;
  Color c;
  GetPixel(x, y, c);
  for (j=0; j<3; j++) c[j] = 255-c[j];
  SetPixel(x, y, c);
}

/* misc prototypes */
unsigned char shadowCalc(int x, int y);
unsigned char shadowCalcSlow(int x, int y);
void shadowPrepare();
void DrawMasses();
void DrawDebug();
void DebugPostprocessPixel(Mass * m, int x, int y, Color color /* variable */, int shadow);

int InitVideo2d(int argc, char **argv)
{
  ResetKeys();

  /* TODO: parse command line */
  
  /* init SDL */ 

  if (SDL_InitSubSystem(SDL_INIT_VIDEO)) {
    printf("Could not initialize the SDL video subsystem: %s\n", SDL_GetError());
    return -1;
  }
  screen = SDL_SetVideoMode(DISPLAY_SCREENX, DISPLAY_SCREENY, DISPLAY_BPP, SDL_SWSURFACE);
  if (!screen) {
    printf("Could not set video mode: %s\n", SDL_GetError());
    return -2;
  }
  SDL_WM_SetCaption("Mass " VERSION, NULL /* TODO: icon */);

  // prepare to draw on the screen
  if ( SDL_MUSTLOCK(screen) ) {
    if ( SDL_LockSurface(screen) < 0 ) {
      printf("Could not lock SDL surface: %s\n", SDL_GetError());
      SDL_QuitSubSystem(SDL_INIT_VIDEO);
      printf("Failed to initialize SDL.\n");
      return -3;
    }
  }

  shadowPrepare();
  DebugLevel(1, "SDL 2D frontend ready.\n");
  return 0;
}

void DoneVideo2d()
{
  if ( SDL_MUSTLOCK(screen) ) SDL_UnlockSurface(screen);
  SDL_QuitSubSystem(SDL_INIT_VIDEO);
  DebugLevel(1, "SDL 2D frontend terminated.\n");
}

void ClearFrame()
{
  SDL_FillRect(screen, NULL, 0);
} 

unsigned char shadowTable[256][256];

unsigned char shadowCalcSlow(int x, int y)
{
  double res, x1, y1;
  int r;

  /* OPTIMIZE with shadow table */
  /* TODO: perhaps move this to effects.c */

  /* HEIGHT_NORM ^= 1 pixel in x or y direction */
  x1 = (double)x/HEIGHT_NORM; y1 = (double)y/HEIGHT_NORM;

  /* calculated for a light direction (1, 1, -1) */
  res = (1/sqrt(x1*x1+y1*y1+1))*(1/sqrt(1+1+1));
  assert(res >= 0);
  res *= (+x1+y1+1);

  res = (res + 1) * 127.5;

  r = res;
  assert(r<=255 && r >= 0);
  //if (r > 255) r = 255;
  //if (r < 0) r = 0;
  return r;
}

/* the maximum height difference - everytihing steeper is displayed all the same */
#define SHADOW_RANGE (HEIGHT_NORM*2)
unsigned char shadowCalc(int x, int y)
{
  x = x * 128 / (SHADOW_RANGE);
  y = y * 128 / (SHADOW_RANGE);
  x += 128;
  y += 128;
  if (x < 0) x = 0;
  if (x > 255) x = 255;
  if (y < 0) y = 0;
  if (y > 255) y = 255;

  /* OPTIMIZE: this operation seems to take long, perhaps a
     recalculation would be faster because it does not have to access
     memory */
  return shadowTable[y][x];
}

void shadowPrepare()
{
  int x, y;
  for (y=-128; y<=127; y++) {
    for (x=-128; x<=127; x++) {
      shadowTable[y+128][x+128] = shadowCalcSlow(x*SHADOW_RANGE/128, y*SHADOW_RANGE/128);
    }
  }
}

void DrawMass(Mass * m);

void DrawMasses()
{
  int i;
  Mass ** visibleMasses;

  visibleMasses = GetMassesInRect(screenRect);
  for (i=0; visibleMasses[i]; i++) DrawMass(visibleMasses[i]);
  free(visibleMasses);
}

void DrawMass(Mass * m)
{
  /* TODO: having the player always in the center of the screen looks
     ugly and boring. But perhaps that's the wrong place to fix this,
     rather change displayCenter. */
  int x, y, i;
  FieldPoint * p;
  Rect visible;	
  /* This assumes (FIELDSIZE > MAX(SCREENX,SCREENY)). No assertion
     here because a small FIELDSIZE is useful for debugging. */
  /* set "visible" to the screen rect, expressed in mass coordinates */
  visible = Wrap_SmallestCommonRect(screenRect, m->maxRect);
  for (y=visible.y; y<visible.y+visible.h; y++) {
    for (x=visible.x; x<visible.x+visible.w; x++) {
      p = FieldXY(x, y);
      if (p->owner == m) {
	/* for each pixel belonging to the mass */
	int shadow = 255; 	/* shuld always get set */
	Color color;
	switch (m->displayType) {
	case DT_NORMAL:
	  /* Calculate shadow value */
	  /* TODO: think about the pixels used here that belong to an other mass */
	  /* OPTIMIZE: this is the performance killer line, as well as
	     the for loop that follows. */
	  shadow = shadowCalc(FieldXY(x+1,y)->height - p->height, FieldXY(x,y+1)->height - p->height);
	  for (i=0; i<3; i++) color[i] = (shadow * m->baseColor[i]) >> 8;
	  /* OK, now follow some if's and but's */
	  if (m->player != -1) {
	    if (p->height < HEIGHT_NORM) {
	      color[0] = color[1] = color[2] = shadow;
	    }
	  }
	  /* anti aliasing (assumes black background color) */
#define HEIGHT_FADEOUT (HEIGHT_NORM*5)
	  if (p->height < HEIGHT_FADEOUT) {
	    for (i=0; i<3; i++) {
	      color[i] -= (HEIGHT_FADEOUT-p->height)*255/(HEIGHT_FADEOUT);
	      if (color[i] < 0) color[i] = 0;
	    }
	  }
	  break;
	case DT_FLAT:
	  memcpy(color, m->baseColor, sizeof(Color));
	  break;
	default:
	  DebugLevel(0, "Warning: I do not know how to draw mass type %d\n", m->displayType);
	  RGBSET(color, 255, 255, 255);
	}

	/* post-process the pixel for debugging */
	if (key.debugHUD.toggled) {
	  DebugPostprocessPixel(m, x, y, color /* variable */, shadow);
	}

	/* now draw the pixel */
	{
	  int sx, sy;
	  sx = (x-screenRect.x)&FIELDWRAP;
	  sy = (y-screenRect.y)&FIELDWRAP;
	  assert(sx >= 0 && sy >= 0);
	  assert(sx < DISPLAY_SCREENX && sy < DISPLAY_SCREENY);
	  SetPixel(sx, sy, color);
	}
	//DebugLevel(3, "SetPixel");
	//SetPixel(x, y, 255, 255, 255);
      }
    }
  }
}

void DrawFrame2d()
{
  if ( SDL_MUSTLOCK(screen) ) SDL_UnlockSurface(screen);

  SDL_FillRect(screen, NULL, 0);

  /* top left corner of the screen in mass coordinates */
  screenRect.x = displayCenter.x - DISPLAY_SCREENX/2;
  screenRect.y = displayCenter.y - DISPLAY_SCREENY/2;

  DrawMasses();
  if (key.debugHUD.toggled) DrawDebug();
  
  SDL_UpdateRect(screen, 0, 0, 0, 0);

  // prepare to draw on the screen
  if ( SDL_MUSTLOCK(screen) ) {
    if ( SDL_LockSurface(screen) < 0 ) {
      printf("Cold not lock screen surface!\n");
      exit(1);
    }
  }
}

void DebugPostprocessPixel(Mass * m, int x, int y, Color color /* variable */, int shadow)
{
  int i;
  FieldPoint * p = FieldXY(x, y);

  switch (key.debugHUDcycle.hits % 4) {
  case 0:
    /* no debug color change */
    break;
  case 1:
    /* HEIGHT_NORM stripes */
    if ((p->height / (HEIGHT_NORM)) % 2 == 0) {
      color[0] = color[1] = color[2] = 255;
    } else {
      color[0] = color[1] = color[2] = 100;
    }	   
    break;
  case 2:
    /* HEIGHT_NORM*4 stripes */
    if ((p->height / (HEIGHT_NORM*3)) % 2 == 0) {
      color[1] = 255;
    } else {
      color[1] = 0;
    }	   
    break;
  case 3:
    /* simple shadow value */
    for (i=0; i<3; i++) color[i] = (m->debugColor[i] * shadow) >> 8;
    break;
  default:
    assert(0);
  }
  if (m == watchedMass) {
    for (i=0; i<3; i++) {
      color[i] += 100;
      if (color[i] > 255) color[i] = 255;
    }
  }
}

void DrawDebug() 
{
  /* mouse in the field */
  Point fmouse;
  fmouse.x = screenRect.x + mouse.pos.x;
  fmouse.y = screenRect.y + mouse.pos.y;

  if (mouse.left.hit) {
    Point p;
    p.x = fmouse.x;
    p.y = fmouse.y;
    watchedMass = FieldPoint(p)->owner;
    /* TODO: print some info about this mass to stdout */
  }

  if (watchedMass) {
    /* make sure the mass still exists, references cannot be relied on */
    if (!IsValidMass(watchedMass)) watchedMass = NULL;
  }

  /* Mass-o-Meter */
  {
    Color bg = {0,0,0}, border = {255,255,255};
    int x, y;
    Rect position;
    position.x = DISPLAY_SCREENX-80-3;
    position.y = 3;
    position.w = 80;
    position.h = 80;

    for (x=position.x; x<position.x+position.w; x++) {
      if (x == position.x || x == position.x+position.w-1) {
	//draw the left / right border
	int y;
	for (y=position.y; y<position.y+position.h; y++) {
	  SetPixel(x, y, border);
	}
      } else {
	Color fg;
	double height;
	int px, py;
	//set (px, py) to the point monitored
	if (mouse.left.toggled) {
	  px = mouse.pos.x;
	  py = mouse.pos.y + (x-position.x-1) - (position.w-2)/2;
	} else {
	  px = mouse.pos.x + (x-position.x-1) - (position.w-2)/2;
	  py = mouse.pos.y;
	}
	if (px < 0 || py < 0 || px >= DISPLAY_SCREENX || py >= DISPLAY_SCREENY) {
	  memcpy(fg, bg, sizeof(fg));
	  height = 0;
	} else {
	  FieldPoint * fp;
	  //first highlight (invert) the point monitored
	  if (!IsPointInRect(XYToPoint(px, py), position) && !key.debugHUDrect.toggled) {
	    InvertPixel(px, py);
	  }
	  //now get height and color
	  fp = FieldXY(px + screenRect.x, py + screenRect.y);
	  if (!fp->owner) {
	    memcpy(fg, bg, sizeof(fg));
	  } else {
	    memcpy(fg, fp->owner->debugColor, sizeof(fg));
	  }
	  height = fp->height;
	  if (height < 0) {
	    //there should never be a negative height value, so mark it red
	    RGBSET(fg, 255, 50, 50);
	    height = 1;
	  } else {
	    int zoom;
	    zoom = key.debugHUDzin.hits - key.debugHUDzout.hits;
	    /* standard display (zoom 0): HEIGHT_NORM ^= 1 pixel */
	    height /= (double)(HEIGHT_NORM);
	    if (zoom > 0) height *= 2 << zoom;
	    if (zoom < 0) height /= 2 << (-zoom);
	  }
	}
	//now display it
	for (y=position.y; y<position.y+position.h; y++) {
	  Color c;
	  if (y == position.y || y == position.y+position.h-1) {
	    memcpy(c, border, sizeof(c));
	  } else if (y == position.y+position.h-2) {
	    memcpy(c, fg, sizeof(c));
	  } else if (y == position.y+position.h-3) {
	    memcpy(c, bg, sizeof(c));
	  } else {
	    int heightPos = (position.h-1 - (y - position.y)) + 3;
	    memcpy(c, (heightPos < height) ? fg : bg, sizeof(c));
	  }
	  SetPixel(x, y, c);
	}
      }
    }
  }

  if (watchedMass) {
    int i;
    Mass * m = watchedMass;
    Rect r;

    for (i=0; i<MAX_SHAPEPOINTS; i++) {
      Point p;
      p = m->shapePoints[i].pos;
      Wrap_FitPointIntoRect(&p, screenRect);
      p.x -= screenRect.x;
      p.y -= screenRect.y;
      InvertPixel(p.x, p.y);
    }

    if (key.debugHUDrect.toggled) {
      //r.x--; r.y--; r.w += 2; r.h += 2;
      r = Wrap_SmallestCommonRect(m->maxRect, screenRect);
      r.x -= screenRect.x;
      r.y -= screenRect.y;
      for (i=r.x; i<r.x+r.w; i++) {
	InvertPixel(i, r.y);
	InvertPixel(i, r.y+r.h-1);
      }
      for (i=r.y; i<r.y+r.h; i++) {
	InvertPixel(r.x, i);
	InvertPixel(r.x+r.w-1, i);
      }
    }
  }
}
